# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import numpy as np
import pandas as pd
import pytest

import mozanalysis.frequentist_stats.bootstrap as mafsb


def test_resample_and_agg_once():
    assert mafsb._resample_and_agg_once(np.array([3.0, 3.0, 3.0]), np.mean) == 3.0


def test_resample_and_agg_once_multistat():
    def _test_run():
        data = np.concatenate([np.zeros(10000), np.ones(10000)])
        res = mafsb._resample_and_agg_once(
            data,
            lambda x: {
                "min": np.min(x),
                "max": np.max(x),
                "mean": np.mean(x),
            },
        )
        return data, res

    for stack_depth in range(4):
        data, res = _test_run()
        assert res["min"] == 0
        assert res["max"] == 1
        assert res["mean"] == pytest.approx(np.mean(data), rel=1e-1)

        if stack_depth >= 3:
            assert res["mean"] != np.mean(data)  # Extremely unlikely
        elif res["mean"] == np.mean(data):
            # This is a 0.5% event - implausible but not impossible.
            # Re-roll the dice a few times to make sure this was a fluke.
            continue


def test_get_bootstrap_samples():
    res = mafsb.get_bootstrap_samples(np.array([3.0, 3.0, 3.0]), num_samples=2)
    assert res.shape == (2,)

    assert res[0] == 3.0
    assert res[1] == 3.0


def test_get_bootstrap_samples_multistat():
    def _test_run():
        data = np.concatenate([np.zeros(10000), np.ones(10000)])
        res = mafsb.get_bootstrap_samples(
            data,
            lambda x: {
                "min": np.min(x),
                "max": np.max(x),
                "mean": np.mean(x),
            },
            num_samples=2,
        )
        return data, res

    for stack_depth in range(4):
        data, res = _test_run()
        assert res.shape == (2, 3)

        assert (res["min"] == 0).all()
        assert (res["max"] == 1).all()
        assert res["mean"].iloc[0] == pytest.approx(np.mean(data), rel=1e-1)
        assert res["mean"].iloc[1] == pytest.approx(np.mean(data), rel=1e-1)

        # If we stuff up (duplicate) the seeds then things aren't random
        assert res["mean"].iloc[0] != res["mean"].iloc[1]

        if stack_depth >= 3:
            assert (res["mean"] != np.mean(data)).any()  # Extremely unlikely
        elif (res["mean"] == np.mean(data)).any():
            # Re-roll the dice a few times to make sure this was a fluke.
            continue


def test_bootstrap_one_branch():
    data = np.concatenate([np.zeros(10000), np.ones(10000)])
    res = mafsb.bootstrap_one_branch(
        data, num_samples=100, summary_quantiles=(0.5, 0.61)
    )

    assert res["mean"] == pytest.approx(0.5, rel=1e-1)
    assert res["0.5"] == pytest.approx(0.5, rel=1e-1)
    assert res["0.61"] == pytest.approx(0.5, rel=1e-1)


def test_bootstrap_one_branch_multistat():
    data = np.concatenate([np.zeros(10000), np.ones(10000), [1e20]])
    res = mafsb.bootstrap_one_branch(
        data,
        stat_fn=lambda x: {
            "max": np.max(x),
            "mean": np.mean(x),
        },
        num_samples=5,
        summary_quantiles=(0.5, 0.61),
        threshold_quantile=0.9999,
    )

    assert res.shape == (2, 3)

    assert res.loc["max", "mean"] == 1
    assert res.loc["max", "0.5"] == 1
    assert res.loc["max", "0.61"] == 1
    assert res.loc["mean", "mean"] == pytest.approx(0.5, rel=1e-1)
    assert res.loc["mean", "0.5"] == pytest.approx(0.5, rel=1e-1)
    assert res.loc["mean", "0.61"] == pytest.approx(0.5, rel=1e-1)


def test_compare_branches():
    data = pd.DataFrame(index=range(60000), columns=["branch", "val"]).astype(
        dtype={"branch": "string", "val": "float"}
    )
    data.iloc[::3, 0] = "control"
    data.iloc[1::3, 0] = "same"
    data.iloc[2::3, 0] = "bigger"

    data.iloc[::2, 1] = 0
    data.iloc[1::2, 1] = 1

    data.iloc[2::12, 1] = 1

    assert data.val[data.branch != "bigger"].mean() == 0.5
    assert data.val[data.branch == "bigger"].mean() == pytest.approx(0.75)

    res = mafsb.compare_branches(data, "val", num_samples=2)

    assert res["individual"]["control"]["mean"] == pytest.approx(0.5, rel=1e-1)
    assert res["individual"]["same"]["mean"] == pytest.approx(0.5, rel=1e-1)
    assert res["individual"]["bigger"]["mean"] == pytest.approx(0.75, rel=1e-1)

    assert "control" not in res["comparative"]
    assert res["comparative"]["same"][("rel_uplift", "exp")] == pytest.approx(
        0, abs=0.1
    )
    assert res["comparative"]["bigger"][("rel_uplift", "exp")] == pytest.approx(
        0.5, abs=0.1
    )

    # num_samples=2 so (0, 0.5, 1):
    assert res["comparative"]["same"][("prob_win", None)] in (0, 0.5, 1)
    assert res["comparative"]["bigger"][("prob_win", None)] == pytest.approx(
        1, abs=0.01
    )


def test_compare_branches_multistat():
    data = pd.DataFrame(index=range(60000), columns=["branch", "val"]).astype(
        dtype={"branch": "string", "val": "float"}
    )
    data.iloc[::3, 0] = "control"
    data.iloc[1::3, 0] = "same"
    data.iloc[2::3, 0] = "bigger"

    data.iloc[::2, 1] = 0
    data.iloc[1::2, 1] = 1

    data.iloc[2::12, 1] = 1

    assert data.val[data.branch != "bigger"].mean() == 0.5
    assert data.val[data.branch == "bigger"].mean() == pytest.approx(0.75)

    res = mafsb.compare_branches(
        data,
        "val",
        stat_fn=lambda x: {
            "max": np.max(x),
            "mean": np.mean(x),
        },
        num_samples=2,
    )

    assert res["individual"]["control"].loc["mean", "mean"] == pytest.approx(
        0.5, rel=1e-1
    )
    assert res["individual"]["same"].loc["mean", "mean"] == pytest.approx(0.5, rel=1e-1)
    assert res["individual"]["bigger"].loc["mean", "mean"] == pytest.approx(
        0.75, rel=1e-1
    )

    assert "control" not in res["comparative"]

    assert res["comparative"]["same"].loc[
        "mean", ("rel_uplift", "exp")
    ] == pytest.approx(0, abs=0.1)
    assert res["comparative"]["bigger"].loc[
        "mean", ("rel_uplift", "exp")
    ] == pytest.approx(0.5, abs=0.1)

    # num_samples=2 so only 3 possible outcomes
    assert res["comparative"]["same"].loc["mean", ("prob_win", None)] in (0, 0.5, 1)
    assert res["comparative"]["bigger"].loc[
        "mean", ("prob_win", None)
    ] == pytest.approx(1, abs=0.01)

    assert res["comparative"]["same"].loc["max", ("rel_uplift", "exp")] == 0
    assert res["comparative"]["bigger"].loc["max", ("rel_uplift", "exp")] == 0


def test_compare_branches_multiple_metrics():
    # Dummy data: Create 2 branches with the same data. Expect no stat sig
    # difference. The denominator values are 10x the numerator values.
    # Expect the statistic to be ~0.1.
    n = 10**3
    df = pd.DataFrame(
        {
            "branch": ["control"] * n + ["treatment"] * n,
            "ad_click": list(range(n, 0, -1)) * 2,
            "sap": [10 * x for x in range(n, 0, -1)] * 2,
        }
    )

    def custom_stat_fn(data):
        return np.sum(data[:, 0]) / np.sum(data[:, 1])

    res = mafsb.compare_branches(
        df,
        ["ad_click", "sap"],
        stat_fn=custom_stat_fn,
        num_samples=50,
        threshold_quantile=0.9,
    )
    assert res["individual"]["control"].loc["mean"] == pytest.approx(0.1, rel=1e-5)
    assert res["individual"]["treatment"].loc["mean"] == pytest.approx(0.1, rel=1e-5)
    assert res["comparative"]["treatment"][("rel_uplift", "exp")] == pytest.approx(
        0, rel=1e-5
    )
    assert res["comparative"]["treatment"][("abs_uplift", "exp")] == pytest.approx(
        0, rel=1e-5
    )


def test_compare_branches_quantiles():
    def _decilize(arr):
        deciles = np.arange(1, 10) * 0.1
        arr_quantiles = np.quantile(arr, deciles)

        arr_dict = {
            f"{label:.1}": arr_quantile
            for label, arr_quantile in zip(deciles, arr_quantiles, strict=False)
        }
        return arr_dict

    N_OBS_PER_BRANCH = 10_000

    data = pd.DataFrame(
        {
            "branch": ["control"] * N_OBS_PER_BRANCH + ["treatment"] * N_OBS_PER_BRANCH,
            "metric_a": np.concatenate(
                [
                    np.random.exponential(scale=1.0, size=N_OBS_PER_BRANCH),
                    np.random.exponential(scale=1.1, size=N_OBS_PER_BRANCH),
                ]
            ),
        }
    )

    results_new = mafsb.compare_branches_quantiles(
        data, "metric_a", quantiles_of_interest=np.arange(1, 10) * 0.1
    )

    results_old = mafsb.compare_branches(data, "metric_a", stat_fn=_decilize)

    for branch in ["control", "treatment"]:
        assert np.isclose(
            results_old["individual"][branch],
            results_new["individual"][branch],
            atol=0.1,
        ).all(), f"branch {branch} individual results differs"

    arr_eq = np.isclose(
        results_old["comparative"]["treatment"]["rel_uplift"],
        results_new["comparative"]["treatment"]["rel_uplift"],
        atol=0.1,
        equal_nan=True,
    )
    assert arr_eq.all(), f"branch {branch} relative differences differ"

    arr_eq = np.isclose(
        results_old["comparative"]["treatment"]["abs_uplift"],
        results_new["comparative"]["treatment"]["abs_uplift"],
        atol=0.1,
        equal_nan=True,
    )
    assert arr_eq.all(), f"branch {branch} absolute differences differ"


def test_get_quantile_bootstrap_samples():
    res = mafsb.get_quantile_bootstrap_samples(
        np.array([3.0, 3.0, 3.0]), quantiles_of_interest=[0.5], num_samples=10
    )
    assert res.shape == (10, 1)

    assert res["0.5"][0] == 3.0
    assert res["0.5"][1] == 3.0



from textwrap import dedent

import pytest
from metric_config_parser import AnalysisUnit

from mozanalysis.experiment import TimeLimits
from mozanalysis.metrics import AnalysisBasis, DataSource, Metric


@pytest.mark.parametrize(
    "experiments_column_type", [None, "simple", "native", "glean", "events_stream"]
)
def test_datasource_constructor_succeeds(experiments_column_type):
    DataSource(
        name="foo",
        from_expr="my_table.name",
        experiments_column_type=experiments_column_type,
    )


@pytest.mark.parametrize(
    "analysis_unit", [AnalysisUnit.CLIENT, AnalysisUnit.PROFILE_GROUP]
)
def test_datasource_build_query_analysis_units(analysis_unit):
    ds = DataSource(
        name="foo",
        from_expr="my_table.name",
        experiments_column_type=None,
    )
    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-14",
        analysis_start_days=0,
        analysis_length_dates=14,
    )
    empty_str = ""
    fddr = tl.first_date_data_required
    lddr = tl.last_date_data_required

    expected_query = f"""SELECT
            e.analysis_id,
            e.branch,
            e.analysis_window_start,
            e.analysis_window_end,
            e.num_exposure_events,
            e.exposure_date,
            {empty_str}
        FROM enrollments e
            LEFT JOIN my_table.name ds
                ON ds.{analysis_unit.value} = e.analysis_id
                AND ds.submission_date BETWEEN '{fddr}' AND '{lddr}'
                AND ds.submission_date BETWEEN
                    DATE_ADD(e.enrollment_date, interval e.analysis_window_start day)
                    AND DATE_ADD(e.enrollment_date, interval e.analysis_window_end day)
                {empty_str}
        GROUP BY
            e.analysis_id,
            e.branch,
            e.num_exposure_events,
            e.exposure_date,
            e.analysis_window_start,
            e.analysis_window_end"""

    query = ds.build_query([], tl, "", None, AnalysisBasis.ENROLLMENTS, analysis_unit)

    assert query == expected_query


@pytest.mark.parametrize(
    ("use_glean_ids", "expected_id"),
    [
        (True, "glean_client_id"),
        (False, "legacy_telemetry_client_id"),
        (None, "client_id"),
    ],
)
def test_datasource_build_query_glean_ids(use_glean_ids, expected_id):
    ds = DataSource(
        name="foo",
        from_expr="my_table.name",
        experiments_column_type=None,
        client_id_column="client_id",
        glean_client_id_column="glean_client_id",
        legacy_client_id_column="legacy_telemetry_client_id",
    )
    tl = TimeLimits.for_single_analysis_window(
        first_enrollment_date="2019-01-01",
        last_date_full_data="2019-01-14",
        analysis_start_days=0,
        analysis_length_dates=14,
    )
    empty_str = ""
    fddr = tl.first_date_data_required
    lddr = tl.last_date_data_required

    expected_query = f"""SELECT
            e.analysis_id,
            e.branch,
            e.analysis_window_start,
            e.analysis_window_end,
            e.num_exposure_events,
            e.exposure_date,
            {empty_str}
        FROM enrollments e
            LEFT JOIN my_table.name ds
                ON ds.{expected_id} = e.analysis_id
                AND ds.submission_date BETWEEN '{fddr}' AND '{lddr}'
                AND ds.submission_date BETWEEN
                    DATE_ADD(e.enrollment_date, interval e.analysis_window_start day)
                    AND DATE_ADD(e.enrollment_date, interval e.analysis_window_end day)
                {empty_str}
        GROUP BY
            e.analysis_id,
            e.branch,
            e.num_exposure_events,
            e.exposure_date,
            e.analysis_window_start,
            e.analysis_window_end"""

    query = ds.build_query(
        [],
        tl,
        "",
        None,
        AnalysisBasis.ENROLLMENTS,
        AnalysisUnit.CLIENT,
        use_glean_ids=use_glean_ids,
    )

    assert query == expected_query


def test_datasource_build_query_union_metric_rows():
    ds = DataSource(
        name="foo",
        from_expr="my_table.name",
        experiments_column_type=None,
    )

    metrics_list = []
    for i in range(2):
        m = Metric(name=f"test_metric_{i}", data_source=ds, select_expr=f"SELECT {i}")
        metrics_list.append(m)

    expected_query = """
                SELECT
                    * EXCEPT (test_metric_0, test_metric_1),
                    'test_metric_0' AS metric_slug,
                    SAFE_CAST(test_metric_0 AS STRING) AS metric_value
                FROM metrics

UNION ALL

                SELECT
                    * EXCEPT (test_metric_0, test_metric_1),
                    'test_metric_1' AS metric_slug,
                    SAFE_CAST(test_metric_1 AS STRING) AS metric_value
                FROM metrics"""

    query = ds.build_query_union_metric_rows(metrics_list, "test-experiment")

    assert expected_query == dedent(query.rstrip())


@pytest.mark.parametrize(
    ("name", "from_expr", "experiments_column_type", "error"),
    [
        (None, "mytable", "simple", TypeError),
        ("name", None, "simple", TypeError),
        ("name", "mytable", "wrong", ValueError),
    ],
)
def test_datasource_constructor_fails(name, from_expr, experiments_column_type, error):
    with pytest.raises(error):
        DataSource(
            name=name,
            from_expr=from_expr,
            experiments_column_type=experiments_column_type,
        )


def test_complains_about_template_without_default():
    with pytest.raises(
        ValueError,
        match="foo: from_expr contains a dataset template but no value was provided.",
    ):
        DataSource(
            name="foo",
            from_expr="moz-fx-data-shared-prod.{dataset}.foo",
        )
    DataSource(
        name="foo",
        from_expr="moz-fx-data-shared-prod.{dataset}.foo",
        default_dataset="dataset",
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from metric_config_parser import AnalysisUnit
from typing_extensions import assert_never

from mozanalysis.types import IncompatibleAnalysisUnit

if TYPE_CHECKING:
    from metric_config_parser.data_source import DataSource as ParserDataSource

    from mozanalysis.experiment import TimeLimits

import logging

import attr

logger = logging.getLogger(__name__)


class AnalysisBasis(Enum):
    """Determines what the population used for the analysis will be based on."""

    ENROLLMENTS = "enrollments"
    EXPOSURES = "exposures"


# attr.s converters aren't compatible with mypy, define our own
# see: https://mypy.readthedocs.io/en/stable/additional_features.html#id1
def client_id_column_converter(client_id_column: str | None) -> str:
    if client_id_column is None:
        return AnalysisUnit.CLIENT.value
    else:
        return client_id_column


def group_id_column_converter(group_id_column: str | None) -> str:
    if group_id_column is None:
        return AnalysisUnit.PROFILE_GROUP.value
    else:
        return group_id_column


def submission_date_column_converter(submission_date_column: str | None) -> str:
    if submission_date_column is None:
        return "submission_date"
    else:
        return submission_date_column


@attr.s(frozen=True, slots=True)
class DataSource:
    """Represents a table or view, from which Metrics may be defined.

    Args:
        name (str): Name for the Data Source. Used in sanity metric
            column names.
        from_expr (str): FROM expression - often just a fully-qualified
            table name. Sometimes a subquery. May contain the string
            ``{dataset}`` which will be replaced with an app-specific
            dataset for Glean apps. If the expression is templated
            on dataset, default_dataset is mandatory.
        experiments_column_type (str or None): Info about the schema
            of the table or view:

            * 'simple': There is an ``experiments`` column, which is an
              (experiment_slug:str -> branch_name:str) map.
            * 'native': There is an ``experiments`` column, which is an
              (experiment_slug:str -> struct) map, where the struct
              contains a ``branch`` field, which is the branch as a
              string.
            * 'glean': There is an ``experiments`` column inside ping_info,
              which is an (experiment_slug:str -> struct) map, where the
              struct contains a ``branch`` field, which is the branch as a
              string.
            * 'events_stream': There is an ``experiment`` within a JSON
              column ``event_extra``. ``branch`` is in the same column.
            * None: There is no ``experiments`` column, so skip the
              sanity checks that rely on it. We'll also be unable to
              filter out pre-enrollment data from day 0 in the
              experiment.
        client_id_column (str, optional): Name of the column that
            contains the ``client_id`` (join key). Defaults to
            'client_id'.
        submission_date_column (str, optional): Name of the column
            that contains the submission date (as a date, not
            timestamp). Defaults to 'submission_date'.
        default_dataset (str, optional): The value to use for
            `{dataset}` in from_expr if a value is not provided
            at runtime. Mandatory if from_expr contains a
            `{dataset}` parameter.
        app_name: (str, optional): app_name used with metric-hub,
            used for validation
        group_id_column (str, optional): Name of the column that
            contains the ``profile_group_id`` (join key). Defaults to
            'profile_group_id'.
        glean_client_id_column (str, optional): Name of the column that
            contains the *glean* telemetry ``client_id`` (join key).
            This is also used to specify that the data source supports glean.
        legacy_client_id_column (str, optional): Name of the column that
            contains the *legacy* telemetry ``client_id`` (join key).
            This is also used to specify that the data source supports legacy.
    """

    name = attr.ib(validator=attr.validators.instance_of(str))
    _from_expr = attr.ib(validator=attr.validators.instance_of(str))
    experiments_column_type = attr.ib(default="simple", type=str | None)
    client_id_column = attr.ib(
        default=AnalysisUnit.CLIENT.value,
        type=str,
        validator=[attr.validators.instance_of(str), attr.validators.min_len(1)],
        converter=client_id_column_converter,
    )
    submission_date_column = attr.ib(
        default="submission_date",
        type=str,
        validator=[attr.validators.instance_of(str), attr.validators.min_len(1)],
        converter=submission_date_column_converter,
    )
    default_dataset = attr.ib(default=None, type=str | None)
    app_name = attr.ib(default=None, type=str | None)
    group_id_column = attr.ib(
        default=AnalysisUnit.PROFILE_GROUP.value,
        type=str,
        validator=[attr.validators.instance_of(str), attr.validators.min_len(1)],
        converter=group_id_column_converter,
    )
    glean_client_id_column = attr.ib(default=None, type=str)
    legacy_client_id_column = attr.ib(default=None, type=str)

    EXPERIMENT_COLUMN_TYPES = (None, "simple", "native", "glean", "events_stream")

    @experiments_column_type.validator
    def _check_experiments_column_type(self, attribute, value):
        if value not in self.EXPERIMENT_COLUMN_TYPES:
            raise ValueError(
                f"experiments_column_type {value!r} must be one of: "
                f"{self.EXPERIMENT_COLUMN_TYPES!r}"
            )

    @default_dataset.validator
    def _check_default_dataset_provided_if_needed(self, attribute, value):
        self.from_expr_for(None)

    def from_expr_for(self, dataset: str | None) -> str:
        """Expands the ``from_expr`` template for the given dataset.
        If ``from_expr`` is not a template, returns ``from_expr``.

        Args:
            dataset (str or None): Dataset name to substitute
                into the from expression.
        """
        effective_dataset = dataset or self.default_dataset
        if effective_dataset is None:
            try:
                return self._from_expr.format()
            except Exception as e:
                raise ValueError(
                    f"{self.name}: from_expr contains a dataset template but no value was provided."  # noqa:E501
                ) from e
        return self._from_expr.format(dataset=effective_dataset)

    @property
    def experiments_column_expr(self) -> str:
        if self.experiments_column_type is None:
            return ""

        elif self.experiments_column_type == "simple":
            return """AND (
                    ds.{submission_date} != e.enrollment_date
                    OR `mozfun.map.get_key`(
                        ds.experiments, '{experiment_slug}'
                    ) IS NOT NULL
                )"""

        elif self.experiments_column_type == "native":
            return """AND (
                    ds.{submission_date} != e.enrollment_date
                    OR `mozfun.map.get_key`(
                        ds.experiments, '{experiment_slug}'
                    ).branch IS NOT NULL
            )"""

        elif self.experiments_column_type == "glean":
            return """AND (
                    ds.{submission_date} != e.enrollment_date
                    OR `mozfun.map.get_key`(
                        ds.ping_info.experiments, '{experiment_slug}'
                    ).branch IS NOT NULL
                )"""

        elif self.experiments_column_type == "events_stream":
            return """AND (
                    ds.{submission_date} != e.enrollment_date
                    OR IF(
                        JSON_VALUE(ds.event_extra, '$.experiment') = '{experiment_slug}',
                        JSON_VALUE(ds.event_extra, '$.branch'),
                        NULL
                    ) IS NOT NULL
                )"""  # noqa:E501

        else:
            raise ValueError

    def build_query(
        self,
        metric_list: list[Metric],
        time_limits: TimeLimits,
        experiment_slug: str,
        from_expr_dataset: str | None = None,
        analysis_basis: AnalysisBasis = AnalysisBasis.ENROLLMENTS,
        analysis_unit: AnalysisUnit = AnalysisUnit.CLIENT,
        exposure_signal=None,
        use_glean_ids: bool | None = None,
    ) -> str:
        """Return a nearly-self contained SQL query.

        This query does not define ``enrollments`` but otherwise could
        be executed to query all metrics from this data source.
        """
        if analysis_unit == AnalysisUnit.CLIENT:
            if use_glean_ids:
                ds_id = self.glean_client_id_column
            elif use_glean_ids is not None:
                ds_id = self.legacy_client_id_column
            else:
                ds_id = self.client_id_column
        elif analysis_unit == AnalysisUnit.PROFILE_GROUP:
            ds_id = self.group_id_column
        else:
            assert_never(analysis_unit)

        if use_glean_ids is not None and not ds_id:
            chosen_id = (
                "glean_client_id_column" if use_glean_ids else "legacy_client_id_column"
            )
            logger.warning(
                f"use_glean_ids set to {use_glean_ids} but {chosen_id} not set."
                f"Falling back to client_id_column {self.client_id_column}"
            )
            ds_id = self.client_id_column

        return """SELECT
            e.analysis_id,
            e.branch,
            e.analysis_window_start,
            e.analysis_window_end,
            e.num_exposure_events,
            e.exposure_date,
            {metrics}
        FROM enrollments e
            LEFT JOIN {from_expr} ds
                ON ds.{ds_id} = e.analysis_id
                AND ds.{submission_date} BETWEEN '{fddr}' AND '{lddr}'
                AND ds.{submission_date} BETWEEN
                    DATE_ADD(e.{date}, interval e.analysis_window_start day)
                    AND DATE_ADD(e.{date}, interval e.analysis_window_end day)
                {ignore_pre_enroll_first_day}
        GROUP BY
            e.analysis_id,
            e.branch,
            e.num_exposure_events,
            e.exposure_date,
            e.analysis_window_start,
            e.analysis_window_end""".format(
            ds_id=ds_id,
            submission_date=self.submission_date_column,
            from_expr=self.from_expr_for(from_expr_dataset),
            fddr=time_limits.first_date_data_required,
            lddr=time_limits.last_date_data_required,
            metrics=",\n            ".join(
                f"{m.select_expr.format(experiment_slug=experiment_slug)} AS {m.name}"
                for m in metric_list
            ),
            date=(
                "exposure_date"
                if analysis_basis == AnalysisBasis.EXPOSURES and exposure_signal is None
                else "enrollment_date"
            ),
            ignore_pre_enroll_first_day=self.experiments_column_expr.format(
                submission_date=self.submission_date_column,
                experiment_slug=experiment_slug,
            ),
        )

    def build_query_union_metric_rows(
        self,
        metric_list: list[Metric],
        experiment_slug: str,
        metrics_query_table: str = "metrics",
    ) -> str:
        """Return a query that produces a unioned and pivoted row-per-metric
        version of the `build_query` results.

        The `metrics_query_table` parameter specifies where the results to be
        pivoted exist. This is expected to be of the format that is returned by
        the query in `build_query`, and can either be a CTE or table/view.

        EXAMPLE:
        --------------------------------------------------------------
        client_id   | enrollment_date   | metric_1  | metric_2
        asdf1234    | 2024-01-01        | True      | 12

        ---- becomes ----

        client_id   | enrollment_date   | metric_slug   | metric_value
        asdf1234    | 2024-01-01        | metric_1      | True
        asdf1234    | 2024-01-01        | metric_2      | 12
        --------------------------------------------------------------
        """

        query_parts = []
        all_metrics = metric_list + self.get_sanity_metrics(experiment_slug)
        metric_names = [m.name for m in all_metrics]
        for metric in all_metrics:
            query_parts.append(f"""
                SELECT
                    * EXCEPT ({", ".join(metric_names)}),
                    '{metric.name}' AS metric_slug,
                    SAFE_CAST({metric.name} AS STRING) AS metric_value
                FROM {metrics_query_table}
            """)

        return "\nUNION ALL\n".join(query_parts)

    def build_query_targets(
        self,
        metric_list: list[Metric],
        time_limits: TimeLimits,
        experiment_name: str,
        analysis_length: int,
        from_expr_dataset: str | None = None,
        continuous_enrollment: bool = False,
        analysis_unit: AnalysisUnit = AnalysisUnit.CLIENT,
    ) -> str:
        """Return a nearly-self contained SQL query that constructs
        the metrics query for targeting historical data without
        an associated experiment slug.

        This query does not define ``targets`` but otherwise could
        be executed to query all metrics from this data source.
        """
        if analysis_unit != AnalysisUnit.CLIENT:
            raise IncompatibleAnalysisUnit(
                "`build_query_targets` currently only supports client_id analysis"
            )

        return """
        SELECT
            t.client_id,
            t.enrollment_date,
            t.analysis_window_start,
            t.analysis_window_end,
            {metrics}
        FROM targets t
            LEFT JOIN {from_expr} ds
                ON ds.{client_id} = t.client_id
                {date_clause}
        GROUP BY
            t.client_id,
            t.enrollment_date,
            t.analysis_window_start,
            t.analysis_window_end""".format(
            client_id=self.client_id_column,
            from_expr=self.from_expr_for(from_expr_dataset),
            metrics=",\n            ".join(
                f"{m.select_expr.format(experiment_name=experiment_name)} AS {m.name}"
                for m in metric_list
            ),
            date_clause=(
                f"""
        AND ds.{self.submission_date_column} BETWEEN '{time_limits.first_date_data_required}' AND '{time_limits.last_date_data_required}'
        AND ds.{self.submission_date_column} BETWEEN
            DATE_ADD(t.enrollment_date, interval t.analysis_window_start day) AND
            DATE_ADD(t.enrollment_date, interval t.analysis_window_end day)"""  # noqa: E501
                if not continuous_enrollment
                else f"""AND ds.{self.submission_date_column} BETWEEN
            t.enrollment_date AND
            DATE_ADD(t.enrollment_date, interval {analysis_length} day)
            """
            ),
        )

    def get_sanity_metrics(self, experiment_slug: str) -> list[Metric]:
        if self.experiments_column_type is None:
            return []

        elif self.experiments_column_type == "simple":
            return [
                Metric(
                    name=self.name + "_has_contradictory_branch",
                    data_source=self,
                    select_expr=agg_any(
                        f"""`mozfun.map.get_key`(
                ds.experiments, '{experiment_slug}'
            ) != e.branch"""
                    ),
                ),
                Metric(
                    name=self.name + "_has_non_enrolled_data",
                    data_source=self,
                    select_expr=agg_any(
                        f"""`mozfun.map.get_key`(
                ds.experiments, '{experiment_slug}'
            ) IS NULL"""
                    ),
                ),
            ]

        elif self.experiments_column_type == "native":
            return [
                Metric(
                    name=self.name + "_has_contradictory_branch",
                    data_source=self,
                    select_expr=agg_any(
                        f"""`mozfun.map.get_key`(
                ds.experiments, '{experiment_slug}'
            ).branch != e.branch"""
                    ),
                ),
                Metric(
                    name=self.name + "_has_non_enrolled_data",
                    data_source=self,
                    select_expr=agg_any(
                        f"""`mozfun.map.get_key`(
                ds.experiments, '{experiment_slug}'
            ).branch IS NULL"""
                    ),
                ),
            ]

        elif self.experiments_column_type == "glean":
            return [
                Metric(
                    name=self.name + "_has_contradictory_branch",
                    data_source=self,
                    select_expr=agg_any(
                        f"""`mozfun.map.get_key`(
                ds.ping_info.experiments, '{experiment_slug}'
            ).branch != e.branch"""
                    ),
                ),
                Metric(
                    name=self.name + "_has_non_enrolled_data",
                    data_source=self,
                    select_expr=agg_any(
                        f"""`mozfun.map.get_key`(
                ds.ping_info.experiments, '{experiment_slug}'
            ).branch IS NULL"""
                    ),
                ),
            ]

        elif self.experiments_column_type == "events_stream":
            return [
                Metric(
                    name=self.name + "_has_contradictory_branch",
                    data_source=self,
                    select_expr=agg_any(
                        f"""IF(
                JSON_VALUE(ds.event_extra, '$.experiment') = '{experiment_slug}',
                JSON_VALUE(ds.event_extra, '$.branch'),
                NULL
            ) != e.branch """
                    ),
                ),
                Metric(
                    name=self.name + "_has_non_enrolled_data",
                    data_source=self,
                    select_expr=agg_any(
                        f"""IF(
                JSON_VALUE(ds.event_extra, '$.experiment') = '{experiment_slug}',
                JSON_VALUE(ds.event_extra, '$.branch'),
                NULL
            ) IS NULL"""
                    ),
                ),
            ]

        else:
            raise ValueError

    @classmethod
    def from_mcp_data_source(
        cls,
        parser_data_source: ParserDataSource,
        app_name: str | None = None,
    ) -> DataSource:
        """metric-config-parser DataSource objects do not have an `app_name`"""
        return cls(
            name=parser_data_source.name,
            from_expr=parser_data_source.from_expression,
            client_id_column=parser_data_source.client_id_column,
            submission_date_column=parser_data_source.submission_date_column,
            experiments_column_type=(
                None
                if parser_data_source.experiments_column_type == "none"
                else parser_data_source.experiments_column_type
            ),
            default_dataset=parser_data_source.default_dataset,
            app_name=app_name,
            group_id_column=parser_data_source.group_id_column,
            glean_client_id_column=parser_data_source.glean_client_id_column,
            legacy_client_id_column=parser_data_source.legacy_client_id_column,
        )


@attr.s(frozen=True, slots=True)
class Metric:
    """Represents an experiment metric.

    Needs to be combined with an analysis window to be measurable!

    Args:
        name (str): A slug; uniquely identifies this metric in tables
        data_source (DataSource): where to find the metric
        select_expr (str): a SQL snippet representing a clause of a SELECT
            expression describing how to compute the metric; must include an
            aggregation function since it will be GROUPed BY the analysis unit
            and branch
        friendly_name (str): A human-readable dashboard title for this metric
        description (str): A paragraph of Markdown-formatted text describing
            what the metric measures, to be shown on dashboards
        app_name: (str, optional): app_name used with metric-hub,
            used for validation
    """

    name = attr.ib(type=str, validator=attr.validators.instance_of(str))
    data_source = attr.ib(
        type=DataSource, validator=attr.validators.instance_of(DataSource)
    )
    select_expr = attr.ib(type=str, validator=attr.validators.instance_of(str))
    friendly_name = attr.ib(type=str | None, default=None)
    description = attr.ib(type=str | None, default=None)
    bigger_is_better = attr.ib(type=bool, default=True)
    app_name = attr.ib(type=str | None, default=None)


def agg_sum(select_expr: str) -> str:
    """Return a SQL fragment for the sum over the data, with 0-filled nulls."""
    return f"COALESCE(SUM({select_expr}), 0)"


def agg_any(select_expr: str) -> str:
    """Return the logical OR, with FALSE-filled nulls."""
    return f"COALESCE(LOGICAL_OR({select_expr}), FALSE)"


def agg_histogram_mean(select_expr: str) -> str:
    """Produces an expression for the mean of an unparsed histogram."""
    return f"""SAFE_DIVIDE(
                SUM(CAST(JSON_EXTRACT_SCALAR({select_expr}, "$.sum") AS int64)),
                SUM((SELECT SUM(value) FROM UNNEST(mozfun.hist.extract({select_expr}).values)))
            )"""  # noqa
